import os
import numpy as np
from topfarm.cost_models.fuga import py_fuga
from topfarm.cost_models.fuga.py_fuga import PyFuga
from topfarm.cost_models.fused_wake_wrappers import FusedWakeGCLWakeModel
from topfarm.cost_models.utils.aep_calculator import AEPCalculator
from topfarm.cost_models.utils.wind_resource import WindResource
from topfarm.plotting import PlotComp
from topfarm.topfarm import TopFarm
from topfarm.cost_models.fuga.lib_reader import read_lib
import matplotlib.pyplot as plt

D = 80.0
D2 = 2 * D
initial_position = np.array([(0, D2), (0, 0), (0, -D2)])
boundary = [(-D2, D2), (D2, D2), (D2, -D2), (-D2, -D2)]
minSpacing = 2.0
fuga_path = os.path.abspath(os.path.dirname(py_fuga.__file__)) + '/Colonel/'


def optimize_AEP_FusedWake_GCL():
    plot_comp = PlotComp()
    f, A, k = read_lib(fuga_path + 'LUT/Farms/Horns Rev 1\hornsrev_north_only_pm45.lib')
    wr = WindResource(f, A, k, ti=np.zeros_like(f) + .1)
    wm = FusedWakeGCLWakeModel(os.path.dirname(__file__) + "/3tb.yml")
    aep_calc = AEPCalculator(wr, wm)
    init_pos = initial_position.copy()
    init_pos[:, 0] += [-20, 0, 20]
    tf = TopFarm(init_pos, aep_calc.get_TopFarm_cost_component(), minSpacing * D, boundary=boundary, plot_comp=plot_comp,
                             driver_options={'optimizer': 'SLSQP'})
    tf.evaluate()
    print(tf.get_cost())
    tf.optimize()
    print(tf.get_cost())
    save_plot('final_gcl.png', tf, False)
    plot_comp.show()


def optimize_AEP_Fuga():
    plot_comp = PlotComp()

    pyFuga = PyFuga(farm_name='Horns Rev 1',
                    turbine_model_path=fuga_path + 'LUT/', turbine_model_name='Vestas_V80_(2_MW_offshore)[h=67.00]',
                    tb_x=initial_position[:, 0], tb_y=initial_position[:, 1],
                    mast_position=(0, 0, 70), z0=0.0001, zi=400, zeta0=0,
                    farms_dir=fuga_path + 'LUT/Farms/', wind_atlas_path='Horns Rev 1\hornsrev_north_only_pm45.lib', climate_interpolation=False)

    init_pos = initial_position.copy()
    init_pos[:, 0] += [-20, 0, 20]
    tf = TopFarm(init_pos, pyFuga.get_TopFarm_cost_component(), minSpacing * D, boundary=boundary, plot_comp=plot_comp,
                 driver_options={'optimizer': 'SLSQP'})
    print (pyFuga.get_aep_gradients())
    save_plot('initial.png', tf, True)
    tf.evaluate()
    print(tf.get_cost())
    tf.optimize()
    print(tf.get_cost())
    save_plot('final_fuga.png', tf, False)
    plot_comp.show()


def save_plot(filename, tf, initial=False):
    return
    import matplotlib.pyplot as plt
    plt.figure(figsize=(3, 3))
    plt.axis('equal')
    plt.xlabel('Distance [m]')
    plt.ylabel('Distance [m]')
    if initial:
        pos = tf.initial_positions
    else:
        pos = tf.turbine_positions
    plt.plot(tf.boundary[:, 0], tf.boundary[:, 1], label='Boundary')
    plt.plot(pos[:, 0], pos[:, 1], '2', ms=15, label='%s turbine positions' % ('Optimized', 'initial')[initial])
    # plt.legend(loc=1)
    plt.tight_layout()
    plt.savefig(filename)
    plt.show()


if __name__ == '__main__':
    optimize_AEP_Fuga()
    optimize_AEP_FusedWake_GCL()



from py_wake.aep_calculator import AEPCalculator
from py_wake.examples.data.iea37._iea37 import IEA37_WindTurbines, IEA37Site
from topfarm._topfarm import TopFarmProblem
from topfarm.constraint_components.boundary import CircleBoundaryConstraint
from topfarm.plotting import XYPlotComp
from topfarm.cost_models.cost_model_wrappers import AEPCostModelComponent
from topfarm.easy_drivers import EasyRandomSearchDriver
from topfarm.drivers.random_search_driver import RandomizeTurbinePosition_Circle
from py_wake.deficit_models.gaussian import IEA37SimpleBastankhahGaussian
import topfarm
import numpy as np
from scipy.interpolate.interpolate import RegularGridInterpolator
import warnings
from py_wake.flow_map import HorizontalGrid
import xarray as xr


class PyWakeAEP(AEPCalculator):
    """TOPFARM wrapper for PyWake AEP calculator"""

    def __init__(self, wake_model):
        warnings.warn('PyWakeAEP is deprecated. Use PyWakeAEPCostModelComponent instead', DeprecationWarning)
        AEPCalculator.__init__(self, wake_model)

    def get_TopFarm_cost_component(self, n_wt, wd=None, ws=None):
        """Create topfarm-style cost component

        Parameters
        ----------
        n_wt : int
            Number of wind turbines
        """
        return AEPCostModelComponent(
            input_keys=['x', 'y'],
            n_wt=n_wt,
            cost_function=lambda **kwargs:
                self.calculate_AEP(x_i=kwargs[topfarm.x_key],
                                   y_i=kwargs[topfarm.y_key],
                                   h_i=kwargs.get(topfarm.z_key, None),
                                   type_i=kwargs.get(topfarm.type_key, 0),
                                   wd=wd, ws=ws).sum(),
            output_unit='GWh')

    def get_aep4smart_start(self, ws=[6, 8, 10, 12, 14], wd=np.arange(360)):
        def aep4smart_start(X, Y, wt_x, wt_y):
            x = np.sort(np.unique(X))
            y = np.sort(np.unique(Y))
            X_j, Y_j, aep_map = self.aep_map(x, y, 0, wt_x, wt_y, wd=wd, ws=ws)
#             import matplotlib.pyplot as plt
#             c = plt.contourf(X_j, Y_j, aep_map[:, :, 0], 100)
#             plt.colorbar(c)
#             plt.show()
            return RegularGridInterpolator((y, x), aep_map.values)(np.array([Y, X]).T)
        return aep4smart_start


class PyWakeAEPCostModelComponent(AEPCostModelComponent):
    def __init__(self, windFarmModel, n_wt, wd=None, ws=None, max_eval=None, **kwargs):
        self.windFarmModel = windFarmModel
        AEPCostModelComponent.__init__(self,
                                       input_keys=[topfarm.x_key, topfarm.y_key],
                                       n_wt=n_wt,
                                       cost_function=lambda **kwargs: self.windFarmModel.aep(
                                           x=kwargs[topfarm.x_key],
                                           y=kwargs[topfarm.y_key],
                                           h=kwargs.get(topfarm.z_key, None),
                                           type=kwargs.get(topfarm.type_key, 0),
                                           wd=wd, ws=ws),
                                       output_unit='GWh',
                                       max_eval=max_eval, **kwargs)

    def get_aep4smart_start(self, ws=[6, 8, 10, 12, 14], wd=np.arange(360)):
        def aep4smart_start(X, Y, wt_x, wt_y, type=0):
            sim_res = self.windFarmModel(wt_x, wt_y, type=type, wd=wd, ws=ws)
            x = np.sort(np.unique(X))
            y = np.sort(np.unique(Y))
            aep_map = sim_res.flow_map(HorizontalGrid(x, y)).aep_xy()
            if isinstance(aep_map, xr.DataArray):
                aep_map = aep_map.values[:, :, 0]
            return RegularGridInterpolator((y, x), aep_map)(np.array([Y, X]).T)
        return aep4smart_start


def main():
    if __name__ == '__main__':
        n_wt = 16
        site = IEA37Site(n_wt)
        windTurbines = IEA37_WindTurbines()
        windFarmModel = IEA37SimpleBastankhahGaussian(site, windTurbines)
        tf = TopFarmProblem(
            design_vars=dict(zip('xy', site.initial_position.T)),
            cost_comp=PyWakeAEPCostModelComponent(windFarmModel, n_wt),
            driver=EasyRandomSearchDriver(randomize_func=RandomizeTurbinePosition_Circle(), max_iter=5),
            constraints=[CircleBoundaryConstraint([0, 0], 1300.1)],
            plot_comp=XYPlotComp())
        tf.optimize()
        tf.plot_comp.show()


main()



"""Objects useful for IEA 37 optimizations
"""
from numpy import array as npa

from py_wake.examples.data.iea37 import iea37_path
from py_wake.examples.data.iea37._iea37 import IEA37Site, IEA37_WindTurbines
from py_wake.examples.data.iea37.iea37_reader import read_iea37_windturbine
from py_wake.deficit_models.gaussian import IEA37SimpleBastankhahGaussian

from topfarm.constraint_components.spacing import SpacingConstraint
from topfarm.constraint_components.boundary import CircleBoundaryConstraint
from topfarm.cost_models.py_wake_wrapper import PyWakeAEPCostModelComponent


def get_iea37_initial(n_wt=9):
    """Initial positions for IEA Task 37 wind farms

    Parameters
    ----------
    n_wt : int, optional
        Number of wind turbines in farm (must be valid IEA 37 farm)
    """
    site = IEA37Site(n_wt)
    return site.initial_position


def get_iea37_constraints(n_wt=9):
    """Constraints for IEA Task 37 wind farms

    Parameters
    ----------
    n_wt : int, optional
        Number of wind turbines in farm (must be valid IEA 37 farm)

    Returns
    -------
    constr : list of topfarm constraints
        Spacing constraint and boundary constraint for IEA 37 model
    """
    diam = read_iea37_windturbine(iea37_path + 'iea37-335mw.yaml')[2]
    spac_constr = SpacingConstraint(2 * diam)
    bound_rad = npa([900, 1300, 2000, 3000])[n_wt == npa([9, 16, 36, 64])][0]
    bound_constr = CircleBoundaryConstraint((0, 0), bound_rad)
    return [spac_constr, bound_constr]


def get_iea37_cost(n_wt=9):
    """Cost component that wraps the IEA 37 AEP calculator"""
    wd = npa(range(16)) * 22.5  # only 16 bins
    site = IEA37Site(n_wt)
    wind_turbines = IEA37_WindTurbines()
    wake_model = IEA37SimpleBastankhahGaussian(site, wind_turbines)
    return PyWakeAEPCostModelComponent(wake_model, n_wt, wd=wd)



class PostConstraint():
    def __init__(self, key, lower=None, upper=None):
        self.key = key
        self.lower = lower
        self.upper = upper

    def __getitem__(self, index):
        return [self.key, self.lower, self.upper][index]

    def __len__(self):
        return 3



from topfarm.tests.test_files.example_data.floris.NREL5MW import NREL5MWREF, ct_curve, power_curve, cp_curve_spline, cp_curve, Amalia1Site
from py_wake.deficit_models.noj import NOJ
from py_wake.aep_calculator import AEPCalculator
import matplotlib.pyplot as plt
import numpy as np
from openmdao.api import view_model, IndepVarComp
from topfarm import TopFarmGroup, TopFarmProblem
from topfarm.constraint_components.boundary import XYBoundaryConstraint
from topfarm.cost_models.cost_model_wrappers import CostModelComponent, AEPCostModelComponent
from topfarm.easy_drivers import EasyRandomSearchDriver
from topfarm.easy_drivers import EasyScipyOptimizeDriver
from topfarm.drivers.random_search_driver import RandomizeTurbinePosition_Square
from topfarm.plotting import XYPlotComp, NoPlot
from topfarm.constraint_components.spacing import SpacingConstraint
#from topfarm.recorders import TopFarmSqliteRecorder
import topfarm
import time


def main():
    if __name__ == '__main__':
        if not 'plantenergy' in topfarm.plugins:
            pass
        else:
            try:

                from plantenergy.GeneralWindFarmGroups import AEPGroup
                from plantenergy.floris import floris_wrapper, add_floris_params_IndepVarComps
                try:
                    import matplotlib.pyplot as plt
                    plt.gcf()
                    plot_comp = XYPlotComp()
                    plot = True
                except RuntimeError:
                    plot_comp = NoPlot()
                    plot = False
    #            plot_comp = NoPlot()

                def setup_prob(differentiable):

                    #####################################
                    ## Setup Floris run with gradients ##
                    #####################################
                    topfarm.x_key = 'turbineX'
                    topfarm.y_key = 'turbineY'
                    turbineX = np.array([1164.7, 947.2, 1682.4, 1464.9, 1982.6, 2200.1])
                    turbineY = np.array([1024.7, 1335.3, 1387.2, 1697.8, 2060.3, 1749.7])
                    f = np.array([3.597152, 3.948682, 5.167395, 7.000154, 8.364547, 6.43485,
                                  8.643194, 11.77051, 15.15757, 14.73792, 10.01205, 5.165975])
                    wind_speed = 8
                    site = Amalia1Site(f, mean_wsp=wind_speed)
                    site.initial_position = np.array([turbineX, turbineY]).T
                    wt = NREL5MWREF()
                    wake_model = NOJ(site, wt)
                    aep_calculator = AEPCalculator(wake_model)
                    n_wt = len(turbineX)
                    differentiable = differentiable
                    wake_model_options = {'nSamples': 0,
                                          'nRotorPoints': 1,
                                          'use_ct_curve': True,
                                          'ct_curve': ct_curve,
                                          'interp_type': 1,
                                          'differentiable': differentiable,
                                          'use_rotor_components': False}

                    aep_comp = AEPGroup(n_wt, differentiable=differentiable, use_rotor_components=False, wake_model=floris_wrapper,
                                        params_IdepVar_func=add_floris_params_IndepVarComps,
                                        wake_model_options=wake_model_options, datasize=len(power_curve), nDirections=len(f), cp_points=len(power_curve))  # , cp_curve_spline=None)

                    def cost_func(AEP, **kwargs):
                        return AEP
                    cost_comp = CostModelComponent(input_keys=[('AEP', [0])], n_wt=n_wt, cost_function=cost_func,
                                                   output_key="aep", output_unit="kWh", objective=True, income_model=True, input_units=['kW*h'])
                    group = TopFarmGroup([aep_comp, cost_comp])
                    boundary = np.array([(900, 1000), (2300, 1000), (2300, 2100), (900, 2100)])  # turbine boundaries
                    prob = TopFarmProblem(
                        design_vars={'turbineX': (turbineX, 'm'), 'turbineY': (turbineY, 'm')},
                        cost_comp=group,
                        driver=EasyRandomSearchDriver(randomize_func=RandomizeTurbinePosition_Square(), max_iter=500),
                        #                        driver=EasyScipyOptimizeDriver(optimizer='SLSQP',tol=10**-12),
                        #                        driver=EasyScipyOptimizeDriver(optimizer='COBYLA'),
                        constraints=[SpacingConstraint(200, units='m'),
                                     XYBoundaryConstraint(boundary, units='m')],
                        plot_comp=plot_comp,
                        expected_cost=-100e2,
                    )
                    turbineZ = np.array([90.0, 100.0, 90.0, 80.0, 70.0, 90.0])
                    air_density = 1.1716    # kg/m^3
                    rotorDiameter = np.zeros(n_wt)
                    hubHeight = np.zeros(n_wt)
                    axialInduction = np.zeros(n_wt)
                    generatorEfficiency = np.zeros(n_wt)
                    yaw = np.zeros(n_wt)
                    for turbI in range(0, n_wt):
                        rotorDiameter[turbI] = wt.diameter()       # m
                        hubHeight[turbI] = wt.hub_height()            # m
                        axialInduction[turbI] = 1.0 / 3.0
                        generatorEfficiency[turbI] = 1.0  # 0.944
                        yaw[turbI] = 0.     # deg.
                    prob['turbineX'] = turbineX
                    prob['turbineY'] = turbineY
                    prob['hubHeight'] = turbineZ
                    prob['yaw0'] = yaw
                    prob['rotorDiameter'] = rotorDiameter
                    prob['hubHeight'] = hubHeight
                    prob['axialInduction'] = axialInduction
                    prob['generatorEfficiency'] = generatorEfficiency
                    prob['windSpeeds'] = np.ones(len(f)) * wind_speed
                    prob['air_density'] = air_density

                    prob['windDirections'] = np.arange(0, 360, 360 / len(f))
                    prob['windFrequencies'] = f / 100
                    # turns off cosine spread (just needs to be very large)
                    prob['model_params:cos_spread'] = 1E12
                    prob['model_params:shearExp'] = 0.25
                    prob['model_params:z_ref'] = 80.
                    prob['model_params:z0'] = 0.
                    prob['rated_power'] = np.ones(n_wt) * 5000.
                    prob['cut_in_speed'] = np.ones(n_wt) * 3
                    prob['cp_curve_wind_speed'] = cp_curve[:, 0]
                    prob['cp_curve_cp'] = cp_curve[:, 1]
                    prob['rated_wind_speed'] = np.ones(n_wt) * 11.4
                    prob['cut_out_speed'] = np.ones(n_wt) * 25.0
                    # if 0:
                    # prob.check_partials(compact_print=True,includes='*direction_group0*')
                    # else:
                    return prob

                differentiable = True
                prob = setup_prob(differentiable)
    #            view_model(prob)
                cost_init, state_init = prob.evaluate()
                tic = time.time()
                cost, state, recorder = prob.optimize()
                toc = time.time()
                print('FLORIS calculation with differentiable = {0} took {1} sec.'.format(differentiable, toc - tic))
                print(prob[topfarm.x_key])

    #            ########################################
    #            ## Setup Floris run without gradients ##
    #            ########################################
    #
    #            differentiable = False
    #            prob2 = setup_prob(differentiable)
    #            cost_init, state_init = prob2.evaluate()
    #            tic = time.time()
    #            cost, state, recorder = prob2.optimize()
    #            toc = time.time()
    #            print('FLORIS calculation with differentiable = {0} took {1} sec.'.format(differentiable, toc-tic))
    #
    #
    #
    #            ########################################
    #            ## Setup Pywake run without gradients ##
    #            ########################################
    #
    #
    #
    #            class PyWakeAEP(AEPCalculator):
    #                """TOPFARM wrapper for PyWake AEP calculator"""
    #
    #                def get_TopFarm_cost_component(self, n_wt, wd=None, ws=None):
    #                    """Create topfarm-style cost component
    #
    #                    Parameters
    #                    ----------
    #                    n_wt : int
    #                        Number of wind turbines
    #                    """
    #                    return AEPCostModelComponent(
    #                        input_keys=['turbineX', 'turbineY'],
    #                        n_wt=n_wt,
    #                        cost_function=lambda **kwargs:
    #                            self.calculate_AEP(x_i=kwargs[topfarm.x_key],
    #                                               y_i=kwargs[topfarm.y_key],
    #                                               h_i=kwargs.get(topfarm.z_key, None),
    #                                               type_i=kwargs.get(topfarm.type_key, None),
    #                                               wd=wd, ws=ws).sum(),
    #                        output_unit='GWh')
    #            aep_calc = PyWakeAEP(site, wt, wake_model)
    #            tf = TopFarmProblem(
    #                    design_vars={'turbineX': turbineX, 'turbineY': turbineY},
    #                cost_comp=aep_calc.get_TopFarm_cost_component(len(turbineX)),
    #                    driver=EasyScipyOptimizeDriver(optimizer='SLSQP'),
    #                    constraints=[SpacingConstraint(200),
    #                                 XYBoundaryConstraint(boundary)],
    #                    plot_comp=plot_comp)
    #            cost_init_pw, state_init_pw = tf.evaluate()
    #            cost_pw, state_pw, recorder_pw = tf.optimize()
    #
    #
    #            prob['turbineX'] = state_pw['turbineX']
    #            prob['turbineY'] = state_pw['turbineY']
    #            cost_pw_fl, state_pw_fl = prob.evaluate()
    #            print('\n***Optimized by PyWake:***')
    #            print('AEP FLORISSE initial:    ', float(-cost_init/10**6))
    #            print('AEP FLORISSE optimized:  ', float(-cost_pw_fl)/10**6)
    #            print('AEP PyWake initial:    ', float(-cost_init_pw))
    #            print('AEP PyWake optimized:  ', float(-cost_pw))
    #
    #            print('***Optimized by FLORISSE:***')
    #            print('AEP FLORISSE initial:    ', float(-cost_init/10**6))
    #            print('AEP FLORISSE optimized:  ', float(-cost)/10**6)
    #            print('AEP Pywake initial:     ', aep_calculator.calculate_AEP(turbineX, turbineY).sum())
    #            print('AEP Pywake optimized:   ', aep_calculator.calculate_AEP(state['turbineX'], state['turbineY']).sum())
            finally:
                topfarm.x_key = 'x'
                topfarm.y_key = 'y'


main()



import os
from fusedwake import fusedwake
import numpy as np
from topfarm.cost_models.fuga import py_fuga
from topfarm.cost_models.fuga.py_fuga import PyFuga
from topfarm.cost_models.fused_wake_wrappers import FusedWakeGCLWakeModel
from topfarm.cost_models.utils.aep_calculator import AEPCalculator
from topfarm.cost_models.utils.wind_resource import WindResource
from topfarm.cost_models.fuga.lib_reader import read_lib

fuga_path = os.path.abspath(os.path.dirname(py_fuga.__file__)) + '/Colonel/'


def HornsrevAEP_FUSEDWake_GCL():
    # wind parameters from "Horns Rev 1\hornsrev2.lib
    wdir_freq, weibull_A, weibull_k = read_lib(fuga_path + 'LUT/Farms/Horns Rev 1\hornsrev2.lib')
    ti = np.zeros_like(wdir_freq) + .1
    wr = WindResource(wdir_freq, weibull_A, weibull_k, ti)
    hornsrev_yml = os.path.dirname(fusedwake.__file__) + "/../examples/hornsrev.yml"
    wm = FusedWakeGCLWakeModel(hornsrev_yml)
    aep_calc = AEPCalculator(wr, wm)
    return aep_calc(wm.windFarm.pos.T)


def HornsrevAEP_Fuga():
    hornsrev_yml = os.path.dirname(fusedwake.__file__) + "/../examples/hornsrev.yml"
    wm = FusedWakeGCLWakeModel(hornsrev_yml)
    tb_x, tb_y = wm.windFarm.pos
    pyFuga = PyFuga(farm_name='Horns Rev 1',
                    turbine_model_path=fuga_path + 'LUT/', turbine_model_name='Vestas_V80_(2_MW_offshore)[h=67.00]',
                    tb_x=tb_x, tb_y=tb_y,
                    mast_position=(0, 0, 70), z0=0.0001, zi=400, zeta0=0,
                    farms_dir=fuga_path + 'LUT/Farms/', wind_atlas_path='Horns Rev 1\hornsrev2.lib')
    return pyFuga.get_aep(wm.windFarm.pos.T)[0]


if __name__ == '__main__':
    print(HornsrevAEP_Fuga())
    print(HornsrevAEP_FUSEDWake_GCL())



from openmdao.drivers.scipy_optimizer import ScipyOptimizeDriver
# from openmdao.drivers.genetic_algorithm_driver import SimpleGADriver  # version 2.5.0 has bug see Issue #874
from topfarm.drivers.genetic_algorithm_driver import SimpleGADriver
from topfarm.drivers.random_search_driver import RandomSearchDriver
import sys
import numpy as np
import openmdao
import scipy


class EasyDriverBase():
    expected_cost = 1
    max_iter = None

    def get_desvar_kwargs(self, model, desvar_name, desvar_values):
        if len(desvar_values) == 4:
            kwargs = {'lower': desvar_values[1], 'upper': desvar_values[2]}
        else:
            kwargs = {}
        return kwargs

    @property
    def supports_expected_cost(self):
        return True


class EasyScipyOptimizeDriver(ScipyOptimizeDriver, EasyDriverBase):

    def __init__(self, optimizer='SLSQP', maxiter=200, tol=1e-8, disp=True, **kwargs):
        """
        Parameters
        ----------
        optimizer : {'COBYLA', 'SLSQP'}
            Gradients are only supported by SLSQP
        maxiter : int
            Maximum number of iterations.
        tol : float
            Tolerance for termination. For detailed control, use solver-specific options.
        disp : bool
            Set to False to prevent printing of Scipy convergence messages
        """
        ScipyOptimizeDriver.__init__(self)
        if optimizer == 'IPOPT':
            try:
                from ipopt.ipopt_wrapper import minimize_ipopt
            except ImportError:
                raise ImportError("""Cannot import ipopt wrapper. Please install cyipopt, e.g. via conda
Windows: conda install -c pycalphad cyipopt
Linux/OSX: conda install -c conda-forge cyipopt
                """)

            def fmt_option(v):
                if isinstance(v, str):
                    return v.encode()
                else:
                    return v
            ipopt_options = {k: fmt_option(v) for k, v in kwargs.items()}

            def minimize_ipopt_wrapper(*args, maxiter=200, disp=True, **kwargs):
                from ipopt.ipopt_wrapper import minimize_ipopt
                ipopt_options.update({'max_iter': self.max_iter or maxiter, 'print_level': int(disp)})
                return minimize_ipopt(*args, options=ipopt_options, **kwargs)
            kwargs = {}
            from openmdao.drivers import scipy_optimizer
            for lst in [scipy_optimizer._optimizers, scipy_optimizer._gradient_optimizers, scipy_optimizer._bounds_optimizers,
                        scipy_optimizer._all_optimizers, scipy_optimizer._constraint_optimizers, scipy_optimizer._constraint_grad_optimizers]:
                lst.add(minimize_ipopt_wrapper)
            optimizer = minimize_ipopt_wrapper

        self.options.update({'optimizer': optimizer, 'maxiter': self.max_iter or maxiter, 'tol': tol, 'disp': disp})
        if kwargs:
            self.options.update(kwargs)

    def get_desvar_kwargs(self, model, desvar_name, desvar_values):
        kwargs = super().get_desvar_kwargs(model, desvar_name, desvar_values)
        if self.options['optimizer'] == 'SLSQP' and tuple([int(v) for v in scipy.__version__.split(".")]) < (1, 5, 0):
            # Upper and lower disturbs SLSQP when running with constraints. Add limits as constraints
            model.add_constraint(desvar_name, kwargs.get('lower', None), kwargs.get('upper', None))
            kwargs = {'lower': np.nan, 'upper': np.nan}  # Default +/- sys.float_info.max does not work for SLSQP

            ref0 = 0
            ref1 = 1
            # TODO: Check if the following improves performance
            # if len(desvar_values) == 4:
            #     ref0 = np.min(desvar_values[1])
            #     ref1 = np.max(desvar_values[2])

            kwargs = {'ref0': ref0, 'ref': ref1, 'lower': np.nan, 'upper': np.nan}
        elif openmdao.__version__ != '2.6.0' and self.options['optimizer'] == 'COBYLA':
            # COBYLA does not work with ref-setting in openmdao 2.6.0.
            # See issue on Github: https://github.com/OpenMDAO/OpenMDAO/issues/942
            if len(desvar_values) == 4:
                ref0 = np.min(desvar_values[1])
                ref1 = np.max(desvar_values[2])
                l, u = [lu * (ref1 - ref0) + ref0 for lu in [desvar_values[1], desvar_values[2]]]
                kwargs = {'ref0': ref0, 'ref': ref1, 'lower': l, 'upper': u}
        return kwargs

    @property
    def supports_expected_cost(self):
        return not (openmdao.__version__ == '2.6.0' and self.options['optimizer'] == 'COBYLA')

    def _get_name(self):
        """Override to add str"""
        return "ScipyOptimize_" + str(self.options['optimizer'])


class EasyIPOPTScipyOptimizeDriver(EasyScipyOptimizeDriver):
    def __init__(self, maxiter=200, tol=1e-8, disp=True,
                 max_cpu_time=1e6,  # : Maximum number of CPU seconds.
                 # A limit on CPU seconds that Ipopt can use to solve one problem. If
                 # during the convergence check this limit is exceeded, Ipopt will
                 # terminate with a corresponding error message. The valid range for this
                 # real option is 0 < max_cpu_time and its default value is 10+06.
                 mu_strategy='monotone',  # : Update strategy for barrier parameter.
                 # Determines which barrier parameter update strategy is to be used. The default value for this string option is "monotone".
                 # Possible values:
                 # - monotone: use the monotone (Fiacco-McCormick) strategy
                 # - adaptive: use the adaptive update strategy
                 acceptable_tol=1e-6,  # : "Acceptable" convergence tolerance (relative).
                 # Determines which (scaled) overall optimality error is considered to be
                 # "acceptable". There are two levels of termination criteria. If the usual
                 # "desired" tolerances (see tol, dual_inf_tol etc) are satisfied at an
                 # iteration, the algorithm immediately terminates with a success message.
                 # On the other hand, if the algorithm encounters "acceptable_iter" many
                 # iterations in a row that are considered "acceptable", it will terminate
                 # before the desired convergence tolerance is met. This is useful in cases
                 # where the algorithm might not be able to achieve the "desired" level of
                 # accuracy. The valid range for this real option is 0 < acceptable_tol and
                 # its default value is 10-06.

                 # All options (https://coin-or.github.io/Ipopt/OPTIONS.html) can be specified via kwargs
                 # The argument type must be correct (str, float or int)
                 **kwargs
                 ):
        EasyScipyOptimizeDriver.__init__(self, optimizer='IPOPT', maxiter=self.max_iter or maxiter, tol=tol, disp=disp,
                                         max_cpu_time=float(max_cpu_time),
                                         mu_strategy=mu_strategy,
                                         acceptable_tol=acceptable_tol,
                                         ** kwargs)


try:
    class PyOptSparseMissingDriver(object):
        options = {}

    from openmdao.drivers.pyoptsparse_driver import pyOptSparseDriver
#     Not working:
#     capi_return is NULL
#     Call-back cb_slfunc_in_slsqp__user__routines failed.
#
#     class EasyPyOptSparseSLSQP(pyOptSparseDriver):
#         def __init__(self, maxit=200, acc=1e-6):
#             pyOptSparseDriver.__init__(self)
#             self.options.update({'optimizer': 'SLSQP'})
#             self.opt_settings.update({'MAXIT': maxit, 'ACC': acc})

    class EasyPyOptSparseIPOPT(pyOptSparseDriver):
        def __init__(self, max_iter=200):
            pyOptSparseDriver.__init__(self)
            self.options.update({'optimizer': 'IPOPT'})
            self.opt_settings.update({'linear_solver': 'ma27', 'max_iter': max_iter,
                                      'start_with_resto': 'yes',
                                      'expect_infeasible_problem': 'yes'})

    from pyoptsparse.pyIPOPT.pyIPOPT import pyipoptcore
    if pyipoptcore is None:
        setattr(sys.modules[__name__], 'EasyPyOptSparseIPOPT', PyOptSparseMissingDriver)

    class EasyPyOptSparseSNOPT(pyOptSparseDriver, EasyDriverBase):
        def __init__(self, major_iteration_limit=200, major_feasibility_tolerance=1e-6, major_optimality_tolerance=1e-6,
                     difference_interval=1e-6, function_precision=1e-8,
                     Print_file='SNOPT_print.out', Summary_file='SNOPT_summary.out', print_results=False):
            """For information about the arguments see
            https://web.stanford.edu/group/SOL/software/snoptHelp/whgdata/whlstt9.htm#9
            """
            pyOptSparseDriver.__init__(self)
            self.options.update({'optimizer': 'SNOPT', 'print_results': print_results})
            self.opt_settings.update({
                'Major feasibility tolerance': major_feasibility_tolerance,
                'Major optimality tolerance': major_optimality_tolerance,
                'Difference interval': difference_interval,
                'Hessian full memory': None,
                'Function precision': function_precision,
                'Major iterations limit': self.max_iter or major_iteration_limit,
                'Print file': Print_file,
                'Summary file': Summary_file,
                'Major step limit': 2.0})

        def get_desvar_kwargs(self, model, desvar_name, desvar_values):
            kwargs = EasyDriverBase.get_desvar_kwargs(self, model, desvar_name, desvar_values)
            if len(desvar_values) == 4:
                ref0 = np.min(desvar_values[1])
                ref1 = np.max(desvar_values[2])
                l, u = desvar_values[1], desvar_values[2]
                kwargs = {'ref0': ref0, 'ref': ref1, 'lower': l, 'upper': u}
            return kwargs


except ModuleNotFoundError:
    for n in ['EasyPyOptSparseSLSQP', 'EasyPyOptSparseIPOPT', 'EasyPyOptSparseSNOPT']:
        setattr(sys.modules[__name__], n, PyOptSparseMissingDriver)


class EasySimpleGADriver(SimpleGADriver, EasyDriverBase):
    def __init__(self, max_gen=100, pop_size=25, Pm=None, Pc=.5, elitism=True,
                 bits={}, debug_print=[], run_parallel=False, random_state=None):
        """SimpleGA driver with optional arguments

        Parameters
        ----------
        max_gen : int
            Number of generations before termination.
        pop_size : int
            Number of points in the GA.
        pm : float
            Probability of mutation.
        pc : float
             Probability of cross over.
        elitism : bool, optional
            If True, replace worst performing point with best from previous generation each iteration.
        bits : dict, optional
            Number of bits of resolution. Default is an empty dict, where every
            unspecified variable is assumed to be integer, and the number of
            bits is calculated automatically.
            If you have a continuous var, you should set a bits value as a key
            in this dictionary.
            Ex: {'x':16,'y':16}
        debug_print : list, optional
            List of what type of Driver variables to print at each iteration. Valid items in list are ‘desvars’,’ln_cons’,’nl_cons’,’objs’
        run_parallel : bool
            Set to True to execute the points in a generation in parallel.
        """
        SimpleGADriver.__init__(self, max_gen=self.max_iter or max_gen, pop_size=pop_size, Pm=Pm, Pc=Pc, elitism=elitism,
                                bits=bits, debug_print=debug_print, run_parallel=run_parallel)
        self.supports['inequality_constraints'] = False
        self.supports['equality_constraints'] = False
        if random_state is not None:
            self._randomstate = random_state


class EasyRandomSearchDriver(RandomSearchDriver, EasyDriverBase):
    def __init__(self, randomize_func, max_iter=100, max_time=600, disp=False, run_parallel=False):
        """Easy initialization of RandomSearchDriver

        Parameters
        ----------
        randomize_func : f(desvar_dict)
            Function to randomize desired variables of desvar_dict
        max_iter : int, optional
            Maximum iterations
        max_time : int, optional
            Maximum time in seconds
        disp : bool
        """
        RandomSearchDriver.__init__(self, randomize_func=randomize_func,
                                    max_iter=self.max_iter or max_iter, max_time=max_time, disp=disp, run_parallel=run_parallel)



import numpy as np
import matplotlib.pyplot as plt
from py_wake.examples.data.hornsrev1 import wt_x, wt_y
from topfarm.tests import npt
from topfarm.cost_models.utils.spanning_tree import mst, spanning_tree


def test_spanning_tree():
    x9, y9 = np.array([[50, 90],
                       [100, 150],
                       [200, 150],
                       [300, 140],
                       [340, 110],
                       [300, 55],
                       [200, 50],
                       [100, 50],
                       [160, 110]]).T

    for x, y, ref in [(x9, y9, 561.0566071630595), (wt_x, wt_y, 44232.604068779656)]:
        x = np.asarray(x)
        y = np.asarray(y)
        sp1 = mst(x, y)
        sp2 = spanning_tree(x, y)
        if 0:
            for sp in [sp1, sp2]:
                plt.figure()
                plt.plot(x, y, '.')
                s = []
                for i in sp.keys():
                    if i[0] < i[1] and i[::-1] in sp:
                        continue
                    plt.plot(x[np.array(i)], y[np.array(i)], 'k-')
                    s.append(np.hypot(np.diff(x[np.array(i)]), np.diff(y[np.array(i)])))
                plt.title("%s, %s" % (sum(sp.values()), np.sum(s)))

            plt.show()
        npt.assert_almost_equal(sum(sp1.values()), ref)
        npt.assert_almost_equal(sum([v for k, v in sp2.items() if k[::-1] not in sp2 or k[0] < k[1]]), ref)



from openmdao.core.explicitcomponent import ExplicitComponent
from openmdao.drivers.genetic_algorithm_driver import SimpleGADriver
import numpy as np


class PenaltyComponent(ExplicitComponent):
    def __init__(self, constraint_comps, as_penalty):
        super().__init__()
        self.post_constraint_lst = constraint_comps
        self.as_penalty = as_penalty

    def setup(self):
        for comp in self.post_constraint_lst:
            self.add_input('penalty_' + comp.const_id, val=0.0)
        self.add_output('penalty', val=0.0)

    def compute(self, inputs, outputs):
        penalties = np.array([inputs[x] for x in inputs])
        if self.as_penalty:
            penalty = sum(penalties)
        else:
            penalty = 0.0
        outputs['penalty'] = penalty


class PostPenaltyComponent(ExplicitComponent):
    def __init__(self, post_constraint_lst, as_penalty):
        super().__init__()
        self.post_constraint_lst = post_constraint_lst
        self.as_penalty = as_penalty

    def setup(self):
        for comp in self.post_constraint_lst:
            self.add_input(comp[0], val=np.zeros(max([len(np.atleast_1d(c)) for c in comp[1:]])))
        self.add_output('post_penalty', val=0.0)

    def compute(self, inputs, outputs):
        def get_penalty(post_constraint):
            key = post_constraint[0]
            if key.startswith('post_penalty'):
                return inputs[key]
            else:
                lower, upper = post_constraint[1:]
                pen = 0
                if lower is not None:
                    pen = max(pen, np.max(lower - inputs[key]))
                if upper is not None:
                    pen = max(pen, np.max(inputs[key] - upper))
                return pen
        if self.as_penalty:
            penalty = np.sum([get_penalty(pc) for pc in self.post_constraint_lst])
        else:
            penalty = 0.0
        outputs['post_penalty'] = penalty



import numpy as np
from openmdao.api import view_model
from py_wake.examples.data.iea37._iea37 import IEA37_WindTurbines, IEA37Site
from py_wake.aep_calculator import AEPCalculator
from topfarm.cost_models.economic_models.turbine_cost import economic_evaluation
from topfarm.cost_models.cost_model_wrappers import CostModelComponent
from topfarm import TopFarmGroup, TopFarmProblem
from topfarm.easy_drivers import EasyRandomSearchDriver
from topfarm.drivers.random_search_driver import RandomizeTurbinePosition_Circle
from topfarm.constraint_components.boundary import CircleBoundaryConstraint
from topfarm.plotting import XYPlotComp, NoPlot
from topfarm.constraint_components.spacing import SpacingConstraint
from py_wake.deficit_models.gaussian import IEA37SimpleBastankhahGaussian
from topfarm.cost_models.py_wake_wrapper import PyWakeAEPCostModelComponent


def main():
    if __name__ == '__main__':
        try:
            import matplotlib.pyplot as plt
            plt.gcf()
            plot_comp = XYPlotComp()
            plot = True
        except RuntimeError:
            plot_comp = NoPlot()
            plot = False

        n_wt = 16
        site = IEA37Site(n_wt)
        windTurbines = IEA37_WindTurbines()
        windFarmModel = IEA37SimpleBastankhahGaussian(site, windTurbines)
        Drotor_vector = [windTurbines.diameter()] * n_wt
        power_rated_vector = [float(windTurbines.power(20) / 1000)] * n_wt
        hub_height_vector = [windTurbines.hub_height()] * n_wt

        def irr_func(aep, **kwargs):
            my_irr = economic_evaluation(Drotor_vector, power_rated_vector, hub_height_vector, aep).calculate_irr()
            print(my_irr)
            return my_irr

        aep_comp = CostModelComponent(
            input_keys=['x', 'y'],
            n_wt=n_wt,
            cost_function=lambda x, y, **_: windFarmModel(x=x, y=y).aep().sum(['wd', 'ws']) * 10**6,
            output_key="aep",
            output_unit="GWh",
            objective=False,
            output_val=np.zeros(n_wt))
        irr_comp = CostModelComponent(
            input_keys=['aep'],
            n_wt=n_wt,
            cost_function=irr_func,
            output_key="irr",
            output_unit="%",
            objective=True,
            income_model=True)
        group = TopFarmGroup([aep_comp, irr_comp])
        problem = TopFarmProblem(
            design_vars=dict(zip('xy', site.initial_position.T)),
            cost_comp=group,
            driver=EasyRandomSearchDriver(randomize_func=RandomizeTurbinePosition_Circle(), max_iter=5),
            constraints=[SpacingConstraint(200),
                         CircleBoundaryConstraint([0, 0], 1300.1)],
            plot_comp=plot_comp)
        cost, state, recorder = problem.optimize()
        #        problem.evaluate()
#        view_model(problem, outfile='example_4_integrated_optimization_aep_and_irr_n2.html', show_browser=False)


main()



from topfarm.utils import smart_start
import numpy as np
from topfarm.tests import npt
from topfarm import TopFarmProblem

from py_wake.examples.data import hornsrev1
from py_wake.deficit_models.noj import NOJ
from topfarm.easy_drivers import EasyScipyOptimizeDriver
from topfarm.constraint_components.boundary import CircleBoundaryConstraint
from topfarm.constraint_components.spacing import SpacingConstraint
from topfarm.cost_models.py_wake_wrapper import PyWakeAEPCostModelComponent, PyWakeAEP
from py_wake.examples.data.iea37._iea37 import IEA37Site

from py_wake.site._site import UniformSite
import pytest


def egg_tray_map():
    x = np.arange(0, 20, 0.1)
    y = np.arange(0, 10, 0.1)
    YY, XX = np.meshgrid(y, x)
    val = np.sin(XX) + np.sin(YY)
    return XX, YY, val


def tests_smart_start():
    xs_ref = [1.6, 14.1, 1.6, 7.9, 14.1, 7.9, 19.9, 19.9, 7.8, 5.8, 14.2,
              5.8, 1.5, 16.2, 16.2, 1.6, 3.7, 14.1, 7.9, 3.7]
    ys_ref = [1.6, 1.6, 7.9, 1.6, 7.9, 7.9, 1.6, 7.9, 5.8, 7.8, 5.8, 1.5, 5.8, 7.8, 1.5, 3.7, 1.6, 3.7, 3.7, 7.9]
    N_WT = 20
    min_space = 2.1

    XX, YY, val = egg_tray_map()
    xs, ys = smart_start(XX, YY, val, N_WT, min_space, seed=0)

    if 0:
        import matplotlib.pyplot as plt
        plt.contourf(XX, YY, val, 100)
        for i in range(N_WT):
            circle = plt.Circle((xs[i], ys[i]), min_space / 2, color='b', fill=False)
            plt.gcf().gca().add_artist(circle)
            plt.plot(xs[i], ys[i], 'rx')
        print(np.round(xs, 1).tolist())
        print(np.round(ys, 1).tolist())

        plt.axis('equal')
        plt.show()
    npt.assert_array_almost_equal([xs, ys], [xs_ref, ys_ref])


def tests_smart_start_random():
    xs_ref = [1.7, 13.9, 1.4, 7.7, 14.4, 7.6, 19.7, 19.7, 8.7, 19.4, 15.8,
              12.4, 7.7, 9.8, 14.1, 1.8, 9.7, 6.6, 13.6, 3.5]
    ys_ref = [7.9, 1.4, 1.7, 1.3, 7.9, 8.4, 1.7, 8.7, 6.4, 6.6, 2.3, 7.1, 3.5, 1.6, 5.8, 5.8, 8.3, 6.5, 3.5, 1.4]

    N_WT = 20
    min_space = 2.1
    XX, YY, val = egg_tray_map()
    np.random.seed(0)

    with pytest.raises(expected_exception=AssertionError):
        xs, ys = smart_start(XX, YY, val, N_WT, min_space, random_pct=101, seed=0)
    xs, ys = smart_start(XX, YY, val, N_WT, min_space, random_pct=1, seed=0)

    if 0:
        import matplotlib.pyplot as plt
        plt.contourf(XX, YY, val, 100)
        plt.plot(XX, YY, ',k')
        for i in range(N_WT):
            circle = plt.Circle((xs[i], ys[i]), min_space / 2, color='b', fill=False)
            plt.gcf().gca().add_artist(circle)
            plt.plot(xs[i], ys[i], 'rx')
        print(np.round(xs, 1).tolist())
        print(np.round(ys, 1).tolist())

        plt.axis('equal')
        plt.show()
    npt.assert_array_almost_equal([xs, ys], [xs_ref, ys_ref])


def tests_smart_start_no_feasible():
    XX, YY, val = egg_tray_map()
    N_WT = 20
    min_space = 5.1

    with pytest.raises(Exception, match="No feasible positions for wt 8"):
        xs, ys = smart_start(XX, YY, val, N_WT, min_space)


@pytest.mark.parametrize('seed,radius,resolution,tol', [(1, 500, 10, 5),
                                                        (0, 2000, 100, 3)])
def test_smart_start_aep_map(seed, radius, resolution, tol):
    site = IEA37Site(16)
    n_wt = 4
    x, y = site.initial_position[:n_wt].T
    wd_lst = np.arange(0, 360, 45)
    ws_lst = [10]
    turbines = hornsrev1.HornsrevV80()
    site = UniformSite([1], .75)
    site.default_ws = ws_lst
    site.default_wd = wd_lst
    wfm = NOJ(site, turbines)
    aep_comp = PyWakeAEPCostModelComponent(wfm, n_wt=n_wt)
    aep_1wt = wfm([0], [0]).aep().sum()

    tf = TopFarmProblem(
        design_vars={'x': x, 'y': y},
        cost_comp=aep_comp,
        driver=EasyScipyOptimizeDriver(),
        constraints=[SpacingConstraint(160), CircleBoundaryConstraint((0, 0), radius)]
    )
    x = np.arange(-radius, radius, resolution)
    y = np.arange(-radius, radius, resolution)
    XX, YY = np.meshgrid(x, y)

    tf.smart_start(XX, YY, aep_comp.get_aep4smart_start(wd=wd_lst, ws=ws_lst), radius=40, plot=0, seed=seed)
    tf.evaluate()

    if 0:
        wt_x, wt_y = tf['x'], tf['y']
        for i, _ in enumerate(wt_x, 1):
            print(wfm(wt_x[:i], wt_y[:i]).aep().sum(['wd', 'ws']))
        aep_comp.windFarmModel(wt_x, wt_y, ws=ws_lst, wd=wd_lst).flow_map().aep_xy().plot()
        print(tf.evaluate())
        import matplotlib.pyplot as plt
        plt.plot(wt_x, wt_y, '2r')
        for c in tf.model.constraint_components:
            c.plot()
        plt.axis('equal')
        plt.show()
    npt.assert_almost_equal(aep_1wt * n_wt, tf['AEP'], tol)


def test_smart_start_aep_map_PyWakeAEP():
    site = IEA37Site(16)
    n_wt = 4

    x, y = site.initial_position[:n_wt].T
    wd_lst = np.arange(0, 360, 45)
    ws_lst = [10]
    turbines = hornsrev1.HornsrevV80()
    site = UniformSite([1], .75)
    site.default_ws = ws_lst
    site.default_wd = wd_lst

    aep = PyWakeAEP(wake_model=NOJ(site, turbines))
    aep_1wt = aep.calculate_AEP([0], [0]).sum()

    tf = TopFarmProblem(
        design_vars={'x': x, 'y': y},
        cost_comp=aep.get_TopFarm_cost_component(n_wt),
        driver=EasyScipyOptimizeDriver(),
        constraints=[SpacingConstraint(160), CircleBoundaryConstraint((0, 0), 500)]
    )
    x = np.arange(-500, 500, 10)
    y = np.arange(-500, 500, 10)
    XX, YY = np.meshgrid(x, y)

    tf.smart_start(XX, YY, aep.get_aep4smart_start(wd=wd_lst, ws=ws_lst), radius=40, seed=1)
    tf.evaluate()

    if 0:
        wt_x, wt_y = tf['x'], tf['y']
        for i, _ in enumerate(wt_x, 1):
            print(aep.calculate_AEP(wt_x[:i], wt_y[:i]).sum((1, 2)))
        X_j, Y_j, aep_map = aep.aep_map(x, y, 0, wt_x, wt_y, ws=ws_lst, wd=wd_lst)
        print(tf.evaluate())
        import matplotlib.pyplot as plt
        c = plt.contourf(X_j, Y_j, aep_map, 100)
        plt.colorbar(c)
        plt.plot(wt_x, wt_y, '2r')
        for c in tf.model.constraint_components:
            c.plot()
        plt.axis('equal')
        plt.show()
    npt.assert_almost_equal(aep_1wt * n_wt, tf['AEP'], 5)



import numpy as np
from py_wake.deficit_models.gaussian import IEA37SimpleBastankhahGaussian
from py_wake.aep_calculator import AEPCalculator
from topfarm.cost_models.economic_models.turbine_cost import economic_evaluation
from topfarm.cost_models.cost_model_wrappers import CostModelComponent
from topfarm import TopFarmGroup, TopFarmProblem
from topfarm.easy_drivers import EasyRandomSearchDriver
from topfarm.drivers.random_search_driver import RandomizeTurbinePosition_Circle, RandomSearchDriver

from topfarm.plotting import XYPlotComp, mypause, NoPlot
from topfarm.constraint_components.spacing import SpacingConstraint

from py_wake.examples.data.dtu10mw import DTU10MW
from py_wake.site import UniformWeibullSite
from py_wake.site.shear import PowerShear
from topfarm.constraint_components.boundary import XYBoundaryConstraint
from py_wake.tests.check_speed import timeit
from openmdao.core.problem import Problem


def get_site():
    f = [0.035972, 0.039487, 0.051674, 0.070002, 0.083645, 0.064348,
         0.086432, 0.117705, 0.151576, 0.147379, 0.10012, 0.05166]
    A = [9.176929, 9.782334, 9.531809, 9.909545, 10.04269, 9.593921,
         9.584007, 10.51499, 11.39895, 11.68746, 11.63732, 10.08803]
    k = [2.392578, 2.447266, 2.412109, 2.591797, 2.755859, 2.595703,
         2.583984, 2.548828, 2.470703, 2.607422, 2.626953, 2.326172]
    ti = 0.001
    h_ref = 100
    alpha = .1
    site = UniformWeibullSite(f, A, k, ti, shear=PowerShear(h_ref=h_ref, alpha=alpha))
    spacing = 2000
    N = 5
    theta = 76  # deg
    dx = np.tan(np.radians(theta))
    x = np.array([np.linspace(0, (N - 1) * spacing, N) + i * spacing / dx for i in range(N)])
    y = np.array(np.array([N * [i * spacing] for i in range(N)]))
    initial_positions = np.column_stack((x.ravel(), y.ravel()))
    eps = 2000
    delta = 5
    site.boundary = np.array([(0 - delta, 0 - delta),
                              ((N - 1) * spacing + eps, 0 - delta),
                              ((N - 1) * spacing * (1 + 1 / dx) + eps * (1 + np.cos(np.radians(theta))),
                               (N - 1) * spacing + eps * np.sin(np.radians(theta)) - delta),
                              ((N - 1) * spacing / dx + eps * np.cos(np.radians(theta)), (N - 1) * spacing + eps * np.sin(np.radians(theta)))])
    site.initial_position = initial_positions
    return site


def main():
    if __name__ == '__main__':
        plot_comp = XYPlotComp()
        site = get_site()
        n_wt = len(site.initial_position)
        windTurbines = DTU10MW()
        min_spacing = 2 * windTurbines.diameter(0)
        windFarmModel = IEA37SimpleBastankhahGaussian(site, windTurbines)
        Drotor_vector = [windTurbines.diameter()] * n_wt
        power_rated_vector = [float(windTurbines.power(20) / 1000)] * n_wt
        hub_height_vector = [windTurbines.hub_height()] * n_wt

        def aep_func(x, y, **_):
            sim_res = windFarmModel(x, y)
            aep = sim_res.aep()
            return aep.sum(['wd', 'ws']).values * 10**6

        def irr_func(aep, **_):
            return economic_evaluation(Drotor_vector, power_rated_vector, hub_height_vector, aep).calculate_irr()

        aep_comp = CostModelComponent(
            input_keys=['x', 'y'],
            n_wt=n_wt,
            cost_function=aep_func,
            output_key="aep",
            output_unit="GWh",
            objective=False,
            output_val=np.zeros(n_wt))
        irr_comp = CostModelComponent(
            input_keys=['aep'],
            n_wt=n_wt,
            cost_function=irr_func,
            output_key="irr",
            output_unit="%",
            objective=True,
            income_model=True)
        group = TopFarmGroup([aep_comp, irr_comp])
        problem = TopFarmProblem(
            design_vars=dict(zip('xy', site.initial_position.T)),
            cost_comp=group,
            driver=EasyRandomSearchDriver(randomize_func=RandomizeTurbinePosition_Circle(), max_iter=10),
            constraints=[SpacingConstraint(min_spacing),
                         XYBoundaryConstraint(site.boundary), ],
            plot_comp=plot_comp)
        cost, state, recorder = problem.optimize()
        problem.plot_comp.show()


main()



"""Example: optimizing a layout with constraints

This example uses a dummy cost function to optimize turbine types.
"""
import numpy as np
import os
from topfarm.cost_models.dummy import DummyCost
from topfarm._topfarm import TopFarmProblem
from openmdao.drivers.doe_generators import FullFactorialGenerator
from topfarm.plotting import TurbineTypePlotComponent, NoPlot


def main():
    if __name__ == '__main__':
        # ------------------------ INPUTS ------------------------

        # define the conditions for the wind farm
        positions = np.array([[0, 0], [6, 6]])  # initial turbine pos
        optimal_types = np.array([[2], [6]])  # optimal layout

        # ===============================================================================
        # Setup the problem and plotting
        # ===============================================================================

        try:
            import matplotlib.pyplot as plt
            plt.gcf()
            plot_comp = TurbineTypePlotComponent(
                turbine_type_names=["Turbine %d" % i for i in range(5)],
                plot_initial=False,
                delay=0.1, legendloc=0)
            plot = True
        except RuntimeError:
            plot_comp = NoPlot()
            plot = False

        # create the wind farm
        tf = TopFarmProblem(
            design_vars={'type': ([0, 0], 0, 4)},
            cost_comp=DummyCost(optimal_types, ['type']),
            plot_comp=plot_comp,
            driver=FullFactorialGenerator(5),
            ext_vars={'x': positions[:, 0], 'y': positions[:, 1]},
        )

        # ===============================================================================
        # #  Run the optimization
        # ===============================================================================
        state = {}
        cost, state, recorder = tf.optimize(state)

        # ===============================================================================
        # plot and prin the the final, optimal types
        # ===============================================================================
        print(state['type'])
        tf.evaluate(state)

        # save the figure
        if plot:
            folder, file = os.path.split(__file__)
            plt.savefig(folder + "/figures/" + file.replace('.py', '.png'))
            plt.show()


main()



import numpy as np
import matplotlib.pyplot as plt
from numpy import newaxis as na
import multiprocessing
import threading
import time


def smart_start(XX, YY, ZZ, N_WT, min_space, radius=None, random_pct=0, plot=False, seed=None):
    """Selects the a number of gridpoints (N_WT) in the grid defined by x and y,
    where ZZ has the maximum value, while chosen points spacing (min_space)
    is respected.

    Parameters
    ----------
    XX : array_like
        x coordinates
    YY : array_like
        y coordinates
    ZZ : array_like
        Values at (XX, YY) of the desired variable in the
        grid points. This could be e.g. the AEP or wind speed.
    N_WT : integer
        number of wind turbines
    min_space: float
        minimum space between turbines
    random_pct : float
        select by random position of the <random_pct> best points
    plot : boolean
        if True, each step is plotted in new figure

    Returns
    -------
    Positions where the aep or wsp is highest, while
    respecting the minimum spacing requirement.

    Notes
    -----
    XX, YY and ZZ can be 1D or 2D, but must have same size
    """
    assert 0 <= random_pct <= 100
    ZZ_is_func = hasattr(ZZ, '__call__')
    if ZZ_is_func:
        arr = np.array([XX.flatten(), YY.flatten()])
    else:
        arr = np.array([XX.flatten(), YY.flatten(), ZZ.flatten()])

    # set radius to None(faster) if if grid resolution > radius
    if radius is not None and (np.diff(np.sort(np.unique(arr[0]))).min() > radius and
                               (np.diff(np.sort(np.unique(arr[1]))).min() > radius)):
        radius = None
    xs, ys = [], []
    if seed is None:
        seed = np.uint32(int((time.time() - int(time.time())) * 1e8 +
                             multiprocessing.current_process().ident + threading.get_ident()) % (2**31))
        np.random.seed(seed)
        seed = np.random.randint(0, 2**31)
    np.random.seed(seed)
    for i in range(N_WT):
        if arr.shape[1] == 0:
            raise Exception('No feasible positions for wt %d' % i)

        if ZZ_is_func:
            z = ZZ(arr[0], arr[1], xs, ys)
        else:
            z = arr[2]

        if radius is not None:
            # average over the rotor area, i.e. all points within one radius from the point
            x, y = arr
            z = np.array([np.mean(z[ind]) for ind in np.hypot((x - x[:, na]), (y - y[:, na])) < radius])

        if random_pct == 0:
            # pick one of the optimal points
            next_ind = np.random.choice(np.where(z == z.max())[0])
        else:
            # pick one of the <random_pct> best points
            n_random = int(np.round(random_pct / 100 * len(z)))
            next_ind = np.random.choice(np.argsort(z)[-(n_random):])

        x0 = arr[0][next_ind]
        y0 = arr[1][next_ind]
        xs.append(x0)
        ys.append(y0)

        if plot:
            plt.figure()
            c = plt.scatter(arr[0], arr[1], c=z)
            plt.colorbar(c)
            plt.plot(xs, ys, '2k', ms=10)
            plt.plot(xs[-1], ys[-1], '2r', ms=10)
            plt.axis('equal')
            plt.show()

        # Remove all point within min_space from the newly added wt
        index = np.where((arr[0] - x0)**2 + (arr[1] - y0)**2 >= min_space**2)[0]
        arr = arr[:, index]

    return xs, ys


def main():
    if __name__ == '__main__':
        N_WT = 30
        min_space = 2.1

        x = np.arange(0, 20, 0.1)
        y = np.arange(0, 10, 0.1)
        YY, XX = np.meshgrid(y, x)
        val = np.sin(XX) + np.sin(YY)
        min_space = 2.1
        xs, ys = smart_start(XX, YY, val, N_WT, min_space)
        c = plt.contourf(XX, YY, val, 100)
        plt.colorbar(c)
        for i in range(N_WT):
            circle = plt.Circle((xs[i], ys[i]), min_space / 2, color='b', fill=False)
            plt.gcf().gca().add_artist(circle)
            plt.plot(xs[i], ys[i], 'rx')
        plt.axis('equal')
        plt.show()


main()



import numpy as np
import multiprocessing
import threading
from topfarm.tests.test_files import xy3tb
from topfarm.easy_drivers import EasyScipyOptimizeDriver


class ParallelRunner():

    def __init__(self, processes=None):
        """
        Parameters
        ----------
        processes : int or None, optional
            Number of processes passed to multiprocessing.Pool
        """
        self.pool = multiprocessing.Pool(processes)

    def __call__(self, state_lst, seq_runner):
        """Run in parallel

        Parameters
        ----------
        state_lst : list
            List of states (as returned by TopFarmProblem.get_DOE_list()
        seq_runner : function
            function for sequential run. Interface must be:
            def f(lst):
                tf = TopfarmProblem(...)
                return tf.optimize()

        Returns
        -------
        best : (cost, state, recorder)
            best result
        results : [(cost1, state1, recorder1),...]
            all results
        """

        indexes = np.round(np.linspace(0, len(state_lst), self.pool._processes + 1)).astype(np.int)
        seq_lst = [state_lst[i1:i2] for i1, i2 in zip(indexes[:-1], indexes[1:])]

        results = self.pool.map(seq_runner, seq_lst)
        results = [r for r_seq in results for r in r_seq]
        best = results[np.argmin([r[0] for r in results])]
        return best, results


# ===============================================================================
# Example functions
# ===============================================================================
def get_topfarm_problem(id, plot=False):
    # setup topfarm problem with max 3 slsqp iterations
    tf = xy3tb.get_tf(plot=plot, driver=EasyScipyOptimizeDriver(maxiter=3, disp=False))
    # Shuffle position via smartstart, XX and YY is posible starting point coordinates
    tf.smart_start(XX=np.linspace(0, 6, 10), YY=np.linspace(-10, 0, 10), random_pct=100)
    return tf


def seq_runner_example(id_lst):
    print("%d cases executed by thread: %s" % (len(id_lst), threading.current_thread().ident))
    # optimize for all elements in lst
    return [get_topfarm_problem(id).optimize() for id in id_lst]
# ===============================================================================
#
# ===============================================================================


def main():
    if __name__ == '__main__':
        # make a list of 4 ids. Could as well be for sets of inputs
        id_lst = np.arange(4)

        print("Current thread: %s" % threading.get_ident())

        # optimize the 4 cases sequential
        print("\nRun on one processor")
        results = seq_runner_example(id_lst)
        for i, r in zip(id_lst, results):
            print("Id %d: %.2f" % (i, r[0]))

        # optimize the four cases in parallel
        print("\nRun on two processors")
        par_runner = ParallelRunner(2)
        (best_cost, best_state, best_recorder), results = par_runner(id_lst, seq_runner_example)
        for i, r in zip(id_lst, results):
            print("Id %d: %.2f" % (i, r[0]))
        print("Minimum cost: %.2f" % best_cost)


main()



import numpy as np
from py_wake.aep_calculator import AEPCalculator
from topfarm.cost_models.economic_models.turbine_cost import economic_evaluation
from topfarm import TopFarmGroup, TopFarmProblem
from py_wake.examples.data import hornsrev1
from topfarm.easy_drivers import EasyRandomSearchDriver, EasySimpleGADriver
from topfarm.drivers.random_search_driver import RandomizeAllUniform
from topfarm.plotting import TurbineTypePlotComponent, NoPlot
from py_wake.wind_turbines import WindTurbines
from topfarm.cost_models.cost_model_wrappers import CostModelComponent
# from py_wake.wind_turbines import cube_power, dummy_thrust
from py_wake.deficit_models.noj import NOJ
from topfarm.constraint_components.capacity import CapacityConstraint
from py_wake.site._site import UniformWeibullSite
import topfarm

import time

def cube_power(ws_cut_in=3, ws_cut_out=25, ws_rated=12, power_rated=5000):
    def power_func(ws):
        ws = np.asarray(ws)
        power = np.zeros_like(ws, dtype=np.float)
        m = (ws >= ws_cut_in) & (ws < ws_rated)
        power[m] = power_rated * ((ws[m] - ws_cut_in) / (ws_rated - ws_cut_in))**3
        power[(ws >= ws_rated) & (ws <= ws_cut_out)] = power_rated
        return power
    return power_func


def dummy_thrust(ws_cut_in=3, ws_cut_out=25, ws_rated=12, ct_rated=8 / 9):
    # temporary thrust curve fix
    def ct_func(ws):
        ws = np.asarray(ws)
        ct = np.zeros_like(ws, dtype=np.float)
        if ct_rated > 0:
            # ct = np.ones_like(ct)*ct_rated
            m = (ws >= ws_cut_in) & (ws < ws_rated)
            ct[m] = ct_rated
            idx = (ws >= ws_rated) & (ws <= ws_cut_out)
            # second order polynomial fit for above rated
            ct[idx] = np.polyval(np.polyfit([ws_rated, (ws_rated + ws_cut_out) / 2,
                                             ws_cut_out], [ct_rated, 0.4, 0.03], 2), ws[idx])
        return ct
    return ct_func

# ----------- SELECT OBJECTIVE & TURN ON/OFF CONSTRAINT --------------
#        obj = False  # objective AEP (True), IRR (False)
#        max_con_on = False  # max installed capacity constraint ON (True), OFF (False)
def main(obj=False, max_con_on=True):
    if __name__ == '__main__':
        start = time.time()
        try:
            import matplotlib.pyplot as plt
            plt.gcf()
            plot = True
        except RuntimeError:
            plot = False

        # ------ DEFINE WIND TURBINE TYPES, LOCATIONS & STORE METADATA -------
        windTurbines = WindTurbines(names=['Ghost_T1', 'T2'],
                                    diameters=[40, 84],
                                    hub_heights=[70, hornsrev1.HornsrevV80().hub_height()],
                                    ct_funcs=[dummy_thrust(ct_rated=0), hornsrev1.HornsrevV80().ct],
                                    power_funcs=[cube_power(power_rated=0), cube_power(power_rated=3000)],
                                    # hornsrev1.HornsrevV80()._power],
                                    power_unit='kW')
        Drotor_vector = windTurbines._diameters
        power_rated_vec = np.array([pcurv(25) / 1000 for pcurv in windTurbines._power_funcs])
        hub_height_vector = windTurbines._hub_heights

        x, y = np.meshgrid(range(-840, 840, 420), range(-840, 840, 420))
        n_wt = len(x.flatten())
        # initial turbine positions and other independent variables
        ext_vars = {'x': x.flatten(), 'y': y.flatten(), 'obj': obj * 1}

        capconst = []
        if max_con_on:
            capconst = [CapacityConstraint(max_capacity=30.01,
                                           rated_power_array=power_rated_vec)]

        # ---------------- DEFINE SITE & SELECT WAKE MODEL -------------------
    #        site = UniformWeibullSite(p_wd=[50, 50], a=[9, 9], k=[2.3, 2.3], ti=.1, alpha=0, h_ref=100)
        site = UniformWeibullSite(p_wd=[100], a=[9], k=[2.3], ti=.1)
        site.default_ws = [9]  # reduce the number of calculations
        site.default_wd = [0]  # reduce the number of calculations

        wake_model = NOJ(site, windTurbines)

        AEPCalc = AEPCalculator(wake_model)

        # ------------- OUTPUTS AEP PER TURBINE & FARM IRR -------------------
        def aep_func(x, y, type, obj, **kwargs):  # TODO fix type as input change to topfarm turbinetype
            out = AEPCalc.calculate_AEP(x_i=x, y_i=y, type_i=type.astype(int)).sum((1, 2))
            if obj:  # if objective is AEP; output the total Farm_AEP
                out = np.sum(out)
            return out * 10**6

        def irr_func(aep, type, **kwargs):
            idx = type.astype(int)
            return economic_evaluation(Drotor_vector[idx],
                                       power_rated_vec[idx],
                                       hub_height_vector[idx],
                                       aep).calculate_irr()

        # ----- WRAP AEP AND IRR INTO TOPFARM COMPONENTS AND THEN GROUP  -----
        aep_comp = CostModelComponent(
            input_keys=[topfarm.x_key, topfarm.y_key, topfarm.type_key, ('obj', obj)],
            n_wt=n_wt,
            cost_function=aep_func,
            output_key="aep",
            output_unit="GWh",
            objective=obj,
            output_val=np.zeros(n_wt),
            income_model=True)
        comps = [aep_comp]  # AEP component is always in the group
        if not obj:  # if objective is IRR initiate/add irr_comp
            irr_comp = CostModelComponent(
                input_keys=[topfarm.type_key, 'aep'],
                n_wt=n_wt,
                cost_function=irr_func,
                output_key="irr",
                output_unit="%",
                objective=True)
            comps.append(irr_comp)

        group = TopFarmGroup(comps)

        # - INITIATE THE PROBLEM WITH ONLY TURBINE TYPE AS DESIGN VARIABLES -
        tf = TopFarmProblem(design_vars={topfarm.type_key: ([0] * n_wt, 0, len(windTurbines._names) - 1)},
                            cost_comp=group,
                            driver=EasyRandomSearchDriver(
                                randomize_func=RandomizeAllUniform([topfarm.type_key]), max_iter=1),
                            # driver=EasySimpleGADriver(max_gen=2, random_state=1),
                            constraints=capconst,
                            # plot_comp=TurbineTypePlotComponent(windTurbines._names),
                            plot_comp=NoPlot(),
                            ext_vars=ext_vars)

        cost, state, rec = tf.optimize()
        # view_model(problem, outfile='ex5_n2.html', show_browser=False)
        end = time.time()
        print(end - start)
        # %%
        # ------------------- OPTIONAL VISUALIZATION OF WAKES ----------------
        post_visual, save = False, False
        if post_visual:
            #        import matplotlib.pyplot as plt
            for cou, (i, j, k, co, ae) in enumerate(zip(rec['x'], rec['y'], rec['type'], rec['cost'], rec['aep'])):
                AEPCalc.calculate_AEP(x_i=i, y_i=j, type_i=k)
                AEPCalc.plot_wake_map(
                    wt_x=i, wt_y=j, wt_type=k, wd=site.default_wd[0], ws=site.default_ws[0], levels=np.arange(
                        2.5, 12, .1))
                windTurbines.plot(i, j, types=k)
                title = f'IRR: {-np.round(co,2)} %, AEP :  {round(np.sum(ae))} GWh, '
                if "totalcapacity" in rec.keys():
                    title += f'Total Capacity: {rec["totalcapacity"][cou]} MW'
                plt.title(title)
                if save:
                    plt.savefig(r'..\..\..\ima2\obj_AEP_{}_MaxConstraint_{}_{}.png'.format(obj, max_con_on, cou))
                plt.show()


for ii in [True, False]:
    for jj in [True, False]:
        main(obj=ii, max_con_on=jj)



import os
from _notebooks.notebook import Notebook
import pytest
import topfarm
from topfarm.cost_models.cost_model_wrappers import CostModelComponent
from topfarm.drivers.random_search_driver import RandomSearchDriver
from topfarm.easy_drivers import EasyDriverBase
from py_wake.tests.check_speed import timeit


def get_notebooks():
    path = os.path.dirname(topfarm.__file__) + "/../_notebooks/elements/"
    return [Notebook(path + f) for f in [f for f in os.listdir(path) if f.endswith('.ipynb')]]


@pytest.mark.parametrize("notebook", get_notebooks())
def test_notebooks(notebook):
    if os.path.basename(notebook.filename) in ['loads.ipynb', 'roads_and_cables.ipynb']:
        pytest.xfail("Notebook, %s, has known issues" % notebook)
    import matplotlib.pyplot as plt

    def no_show(*args, **kwargs):
        pass
    plt.show = no_show  # disable plt show that requires the user to close the plot
    print(notebook.filename)
    try:
        notebook.check_code()
        notebook.check_links()
    except Exception as e:
        raise Exception(notebook.filename + " failed") from e
    finally:
        plt.close()


if __name__ == '__main__':
    print([os.path.basename(n.filename) for n in get_notebooks()])
